/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include "queen/grid.h"

#include "queen/display.h"
#include "queen/logic.h"
#include "queen/queen.h"
#include "queen/command.h"

#include "common/debug.h"

#include "enhanced/GameSettingsManager.h"

#define MAX_HOTSPOT_RECTS 200
#define MAX_CHAT_RECTS 6

using Common::Rect;

namespace Queen {

struct HotspotTransformation {
	uint16 mRoomNumber;
	Rect mOriginal;
	Rect mNew;
	bool mDeleteHotspot;

	/*	HotspotTransformation(uint16 roomNumber, Rect originalHotspot, Rect newHotspot, bool deleteHotspot)
	 {
	 mRoomNumber = roomNumber;
	 mOriginal = originalHotspot;
	 mNew = newHotspot;
	 mDeleteHotspot = deleteHotspot;
	 }*/
};

const HotspotTransformation hotspotTransformations[] = { ///
		///
				{ 102, { 0, 127, 20, 132 }, { 49, 72, 290, 149 }, false }, // Mosaic room
				{ 102, { 0, 132, 20, 148 }, { 0, 0, 0, 0 }, true }, // Mosaic room
				{ 68, { 278, 8, 284, 20 }, { 0, 0, 0, 0 }, true }, // Dummy hotspot
				{ 68, { 0, 0, 48, 149 }, { 0, 129, 140, 149 }, false }, // Path
				{ 66, { 265, 35, 319, 149 }, { 235, 129, 319, 149 }, false }, // Path
				{ 68, { 281, 29, 319, 149 }, { 0, 0, 0, 0 }, true }, // Large Dinosaur
				{ 68, { 164, 29, 281, 149 }, { 164, 29, 319, 149 }, false }, // Large Dinosaur
				{ 67, { 246, 6, 282, 30 }, { 0, 0, 0, 0 }, true }, // Carcass
				{ 69, { 28, 2, 142, 73 }, { 0, 0, 0, 0 }, true }, // Monster Frank
				{ 69, { 24, 0, 28, 25 }, { 0, 0, 0, 0 }, true } // Zeppelin

		};

const uint16 HOTSPOT_TRANSFORMATIONS_NUM = 10;

Grid::Grid(QueenEngine *vm) :
		_vm(vm) {
	memset(_zones, 0, sizeof(_zones));

	mCurrentHotspots = new Enhanced::Hotspot[MAX_HOTSPOT_RECTS];
	mCurrentHotspotCount = 0;

	mChatHitAreas = new Rect[MAX_CHAT_RECTS];
	mChatHitAreasCount = 0;
}

Grid::~Grid() {
	delete[] _objMax;
	delete[] _areaMax;
	delete[] _area;
	delete[] _objectBox;

	delete[] mCurrentHotspots;
	delete[] mChatHitAreas;
}

void Grid::readDataFrom(uint16 numObjects, uint16 numRooms, byte *&ptr) {
	uint16 i, j;

	_numRoomAreas = numRooms;

	_objMax = new int16[_numRoomAreas + 1];
	_areaMax = new int16[_numRoomAreas + 1];
	_area = new Area[_numRoomAreas + 1][MAX_AREAS_NUMBER];

	_objMax[0] = 0;
	_areaMax[0] = 0;
	memset(&_area[0], 0, sizeof(Area) * MAX_AREAS_NUMBER);
	for (i = 1; i <= _numRoomAreas; i++) {
		_objMax[i] = (int16) READ_BE_INT16(ptr);
		ptr += 2;
		_areaMax[i] = (int16) READ_BE_INT16(ptr);
		ptr += 2;
		memset(&_area[i][0], 0, sizeof(Area));
		for (j = 1; j <= _areaMax[i]; j++) {
			assert(j < MAX_AREAS_NUMBER);
			_area[i][j].readFromBE(ptr);
		}
	}

	_objectBox = new Box[numObjects + 1];
	memset(&_objectBox[0], 0, sizeof(Box));
	for (i = 1; i <= numObjects; i++) {
		_objectBox[i].readFromBE(ptr);
	}
}

void Grid::setZone(GridScreen screen, uint16 zoneNum, uint16 x1, uint16 y1,
		uint16 x2, uint16 y2) {
	LOGD("Grid::setZone(%d, %d, (%d,%d), (%d,%d))", screen, zoneNum, x1, y1, x2,
			y2);
	assert(zoneNum < MAX_ZONES_NUMBER);
	ZoneSlot *pzs = &_zones[screen][zoneNum];
	pzs->valid = true;
	pzs->box.x1 = x1;
	pzs->box.y1 = y1;
	pzs->box.x2 = x2;
	pzs->box.y2 = y2;
}

void Grid::setZone(GridScreen screen, uint16 zoneNum, const Box &box) {
	LOGD("Grid::setZone(%d, %d, (%d,%d), (%d,%d))", screen, zoneNum, box.x1,
			box.y1, box.x2, box.y2);
	assert(zoneNum < MAX_ZONES_NUMBER);
	ZoneSlot *pzs = &_zones[screen][zoneNum];
	pzs->valid = true;
	pzs->box = box;
}

uint16 Grid::findZoneForPos(GridScreen screen, uint16 x, uint16 y) const {
//	LOGD("Logic::findZoneForPos(%d, (%d,%d))", screen, x, y);
	int i;
	if (screen == GS_PANEL) {
		y -= ROOM_ZONE_HEIGHT;
	}
	for (i = 1; i < MAX_ZONES_NUMBER; ++i) {
		const ZoneSlot *pzs = &_zones[screen][i];
		if (pzs->valid && pzs->box.contains(x, y)) {

			// HACK: remove incorrect "Carcass" hotspot
			if (_vm->logic()->currentRoom() == 67 && pzs->box.x1 == 246
					&& pzs->box.y1 == 6 && pzs->box.x2 == 282
					&& pzs->box.y2 == 30) {

				continue;
			}

			// HACK: remove extra "Monster Frank" hotspot
			if (_vm->logic()->currentRoom() == 69 && pzs->box.x1 == 28
					&& pzs->box.y1 == 2 && pzs->box.x2 == 142
					&& pzs->box.y2 == 73
                && Enhanced::GameSettingsManager::instance()->isClassicMode() == false) {

				continue;
			}

			// HACK: modify extra "Path" hotspot
			if (_vm->logic()->currentRoom() == 66 && pzs->box.x1 == 265
					&& pzs->box.y1 == 35 && pzs->box.x2 == 319
					&& pzs->box.y2 == 149
					&& Enhanced::GameSettingsManager::instance()->isClassicMode() == false && y < 129) {

				continue;
			}

			//LOGD("Grid::findZoneForPos: found zone %d %d %d %d", pzs->box.x1, pzs->box.y1, pzs->box.x2, pzs->box.y2);
			return i;
		}
	}
	return 0;
}

uint16 Grid::findAreaForPos(GridScreen screen, uint16 x, uint16 y) const {
	uint16 room = _vm->logic()->currentRoom();
	uint16 zoneNum = findZoneForPos(screen, x, y);
	if (zoneNum <= _objMax[room]) {
		zoneNum = 0;
	} else {
		zoneNum -= _objMax[room];
	}
	return zoneNum;
}

void Grid::clear(GridScreen screen) {
	debug(9, "Grid::clear(%d)", screen);
	for (int i = 1; i < MAX_ZONES_NUMBER; ++i) {
		_zones[screen][i].valid = false;
	}
}

void Grid::setupNewRoom(uint16 room, uint16 firstRoomObjNum) {
	debug(9, "Grid::setupNewRoom()");
	clear(GS_ROOM);

	clearCurrentHotspots();

	uint16 i;
	uint16 zoneNum;

	//mCachedRoom = room;
	//mCachedFirstRoomObjNum = firstRoomObjNum;

	// setup objects zones
	uint16 maxObjRoom = _objMax[room];
	zoneNum = 1;
	for (i = firstRoomObjNum + 1; i <= firstRoomObjNum + maxObjRoom; ++i) {
		if (_vm->logic()->objectData(i)->name != 0) {
			LOGD("Grid::setupNewRoom: setup object zones");
			if (room == 41 && i == 303) {

				// WORKAROUND bug #1599009: In the room 41, the bounding box of the
				// stairs (object 303) doesn't match with the room picture. With the
				// original box dimensions, Joe could walk "above" the stairs, giving
				// the impression of floating in the air.
				// To fix this, the bounding box is set relative to the position of
				// the cabinet (object 295).

				uint16 y1 = _objectBox[295].y2 + 1;
				setZone(GS_ROOM, zoneNum, _objectBox[i].x1, y1,
						_objectBox[i].x2, _objectBox[i].y2);

				addToCurrentHotspots(
						Common::Rect(_objectBox[i].x1, y1, _objectBox[i].x2,
								_objectBox[i].y2));

			} else {
				setZone(GS_ROOM, zoneNum, _objectBox[i]);

				addToCurrentHotspots(
						Common::Rect(_objectBox[i].x1, _objectBox[i].y1,
								_objectBox[i].x2, _objectBox[i].y2));
			}
		} else {
			addToCurrentHotspots(Common::Rect(0, 0, 0, 0));
		}
		++zoneNum;
	}

	mMaxObjectZoneNum = zoneNum;

	// setup room zones (areas)
	uint16 maxAreaRoom = _areaMax[room];
	for (zoneNum = 1; zoneNum <= maxAreaRoom; ++zoneNum) {
		LOGD("Grid::setupNewRoom: setup room zones");
		setZone(GS_ROOM, maxObjRoom + zoneNum, _area[room][zoneNum].box);
	}
}

void Grid::setupPanel() {
	for (int i = 0; i <= 7; ++i) {
		uint16 x = i * 20;
		setZone(GS_PANEL, i + 1, x, 10, x + 19, 49);
	}

	// inventory scrolls
	setZone(GS_PANEL, 9, 160, 10, 179, 29);
	setZone(GS_PANEL, 10, 160, 30, 179, 49);

	// inventory items
	setZone(GS_PANEL, 11, 180, 10, 213, 49);
	setZone(GS_PANEL, 12, 214, 10, 249, 49);
	setZone(GS_PANEL, 13, 250, 10, 284, 49);
	setZone(GS_PANEL, 14, 285, 10, 320, 49);
}

void Grid::drawZones() {
	for (int i = 1; i < MAX_ZONES_NUMBER; ++i) {
		const ZoneSlot *pzs = &_zones[GS_ROOM][i];
		if (pzs->valid) {
			const Box *b = &pzs->box;
			_vm->display()->drawBox(b->x1, b->y1, b->x2, b->y2, 3);
		}
	}
}

const Box *Grid::zone(GridScreen screen, uint16 index) const {
	const ZoneSlot *zs = &_zones[screen][index];
	assert(zs->valid);
	return &zs->box;
}

Verb Grid::findVerbUnderCursor(int16 cursorx, int16 cursory) const {
	static const Verb pv[] = { VERB_NONE, VERB_OPEN, VERB_CLOSE, VERB_MOVE,
			VERB_GIVE, VERB_LOOK_AT, VERB_PICK_UP, VERB_TALK_TO, VERB_USE,
			VERB_SCROLL_UP, VERB_SCROLL_DOWN, VERB_INV_1, VERB_INV_2,
			VERB_INV_3, VERB_INV_4, };
	return pv[findZoneForPos(GS_PANEL, cursorx, cursory)];
}

uint16 Grid::findObjectUnderCursor(int16 cursorx, int16 cursory) const {

	uint16 roomObj = 0;
	if (cursory < ROOM_ZONE_HEIGHT) {
		int16 x = cursorx + _vm->display()->horizontalScroll();
		roomObj = findZoneForPos(GS_ROOM, x, cursory);
	}

	return roomObj;
}

uint16 Grid::findObjectNumber(uint16 zoneNum) const {
	// l.316-327 select.c
	uint16 room = _vm->logic()->currentRoom();
	uint16 obj = zoneNum;
	uint16 objectMax = _objMax[room];
	debug(9, "Grid::findObjectNumber(%X, %X)", zoneNum, objectMax);
	if (zoneNum > objectMax) {
// this is an area box, check for associated object
		obj = _area[room][zoneNum - objectMax].object;
		if (obj != 0) {
			// there is an object, get its number
			obj -= _vm->logic()->currentRoomData();
		}
	}
	return obj;
}

uint16 Grid::findScale(uint16 x, uint16 y) const {
	uint16 room = _vm->logic()->currentRoom();
	uint16 scale = 100;
	uint16 areaNum = findAreaForPos(GS_ROOM, x, y);
	if (areaNum != 0) {
		scale = _area[room][areaNum].calcScale(y);
	}
	return scale;
}

void Grid::saveState(byte *&ptr) {
	uint16 i, j;
	for (i = 1; i <= _numRoomAreas; ++i) {
		for (j = 1; j <= _areaMax[i]; ++j) {
			_area[i][j].writeToBE(ptr);
		}
	}
}

void Grid::loadState(uint32 ver, byte *&ptr) {
	uint16 i, j;
	for (i = 1; i <= _numRoomAreas; ++i) {
		for (j = 1; j <= _areaMax[i]; ++j) {
			_area[i][j].readFromBE(ptr);
		}
	}
}

uint16 Grid::getCurrentHotspots(Enhanced::Hotspot* rectArray) {

	//LOGD("Grid::getCurrentHotspots: room %d", _vm->logic()->currentRoom());

	uint16 hotspotCount = 0;
	for (int i = 1; i < mMaxObjectZoneNum; ++i) {
		const ZoneSlot *pzs = &_zones[GS_ROOM][i];
		if (pzs->valid) {

			int objNum = findObjectNumber(i);
			ObjectData *od = _vm->command()->findObjectData(objNum);
			if (od != NULL && od->name > 0) {

				// Copy the hotspot
				Box currentBox = pzs->box;
				Rect rect = Common::Rect(currentBox.x1, currentBox.y1,
						currentBox.x2, currentBox.y2);

			//	LOGD("Grid::getCurrentHotspots: before scrolling %s",
			//			rect.debugStr());

				// Apply hotspot transformations
				bool shouldDelete = false;
				for (int i = 0; i < HOTSPOT_TRANSFORMATIONS_NUM; ++i) {
					const HotspotTransformation& hotspotTransformation =
							hotspotTransformations[i];
					if (hotspotTransformation.mRoomNumber
							== _vm->logic()->currentRoom()
							&& hotspotTransformation.mOriginal.equals(rect)) {
						if (hotspotTransformation.mDeleteHotspot) {
							LOGD(
									"Grid::addToCurrentHotspots: deleting hotspot");
							shouldDelete = true;
							break;
						} else {
							rect = hotspotTransformation.mNew;
							LOGD(
									"Grid::addToCurrentHotspots: applying transformation - new rect %s",
									rect.debugStr());

							break;
						}
					}
				}

				if (shouldDelete)
					continue;

				// Handle scrolling
				int16 scrollX = _vm->display()->horizontalScroll();
				rect.left -= scrollX;
				rect.right -= scrollX;
				if (rect.left >= GAME_SCREEN_WIDTH || rect.right <= 0) {
					continue;
				}

			//	LOGD("Grid::getCurrentHotspots: name %s rect %s",
			//			_vm->logic()->objectName(od->name), rect.debugStr());

				rectArray[hotspotCount].mRect = rect;

				Verb defaultVerb = State::findDefaultVerb(od->state);
				if (defaultVerb == VERB_NONE) {
					defaultVerb = VERB_WALK_TO;
				}

				rectArray[hotspotCount].mHotspotType = defaultVerb;

				++hotspotCount;
			}
		}
	}

	return hotspotCount;

	/*	uint16 maxObjRoom = _objMax[mCachedRoom];
	 uint16 count = 0;
	 for (int i = 0; i < maxObjRoom; ++i) {
	 int objNum = i + mCachedFirstRoomObjNum + 1;
	 if (_vm->logic()->objectData(objNum)->name > 0) {
	 // Copy the hotspot
	 rectArray[count] = mCurrentHotspots[i];
	 ++count;
	 }
	 }

	 return count;*/
}

void Grid::addToCurrentHotspots(Common::Rect rect) {

	LOGD("Grid::addToCurrentHotspots: %d %d %d %d", rect.left, rect.top,
			rect.right, rect.bottom);

	mCurrentHotspots[mCurrentHotspotCount].mRect = rect;
	++mCurrentHotspotCount;
}

void Grid::clearCurrentHotspots() {

	LOGD("Grid::clearCurrentHotspots: ");

	mCurrentHotspotCount = 0;
}

void Grid::getChatHitAreas(Rect* rectArray, uint16& count) {
	for (int i = 0; i < mChatHitAreasCount; ++i) {
		rectArray[i] = mChatHitAreas[i];
	}

	count = mChatHitAreasCount;
}

void Grid::addToChatHitAreas(Common::Rect rect) {

	LOGD("Grid::addToChatHitAreas: %d %d %d %d", rect.left, rect.top,
			rect.right, rect.bottom);

	mChatHitAreas[mChatHitAreasCount] = rect;
	++mChatHitAreasCount;

}
void Grid::clearChatHitAreas() {
	LOGD("Grid::clearChatHitAreas: ");

	mChatHitAreasCount = 0;
}

} // End of namespace Queen
